/*
 * $Id: inout.c,v 1.1.1.1 2002/08/23 13:52:08 kingst Exp $
 */

#include <linux/assert.h>
#include <asm/lotol.h>


static unsigned long in_res = 0x12345678;	/* random number */

unsigned char inb(unsigned short port)
{
	/* keep list sorted alphabetically! */

	extern unsigned char umapic_inb(unsigned short);
	extern unsigned char umide_inb(int, unsigned short);
	extern unsigned char umkbd_inb(unsigned short);
	extern unsigned char umnet_inb(int, unsigned short);
	extern unsigned char umpic_inb(int, unsigned short);
	extern unsigned char umpit_inb(int, unsigned short);
	extern unsigned char umppi_inb(unsigned short);
	extern unsigned char umrtc_inb(unsigned short);
	extern unsigned char umvga_inb(unsigned short);
	unsigned char value;

	/* Keep list sorted by port number! */

	if (0x0020 <= port && port <= 0x0021) {
		/* pic (master) */
		value = umpic_inb(0, port - 0x0020);

	} else if (0x0022 <= port && port <= 0x23) {
		/* IMCR (Interrupt Mode Configuration Register) */
		value = umapic_inb(port - 0x0022);

	} else if (0x0040 <= port && port <= 0x0043) {
		/* pit0 */
		value = umpit_inb(0, port - 0x0040);

	} else if (0x0048 <= port && port <= 0x004b) {
		/* pit1 */
		value = umpit_inb(1, port - 0x0048);

	} else if (0x0060 == port || port == 0x0064) {
		/* keyboard */
		value = umkbd_inb(port);

	} else if (0x0061 <= port && port <= 0x0063) {
		/* PPI 8255 (multi purpose I/O) */
		value = umppi_inb(port - 0x0061);

	} else if (0x0070 <= port && port <= 0x007f) {
		/* real time clock */
		value = umrtc_inb(port);

	} else if (0x00a0 <= port && port <= 0x00a1) {
		/* pic (slave) */
		value = umpic_inb(1, port - 0x00a0);

	} else if (0x0160 <= port && port < 0x0168) {
		/* ide5 reg0-7 */
		value = umide_inb(5, port - 0x0160);

	} else if (0x0168 <= port && port < 0x0170) {
		/* ide3 reg0-7 */
		value = umide_inb(3, port - 0x0168);

	} else if (0x0170 <= port && port < 0x0178) {
		/* ide1 reg0-7 */
		value = umide_inb(1, port - 0x0170);

	} else if (0x01e0 <= port && port < 0x01e8) {
		/* ide4 reg0-7 */
		value = umide_inb(4, port - 0x01e0);

	} else if (0x01e8 <= port && port < 0x01f0) {
		/* ide2 reg0-7 */
		value = umide_inb(2, port - 0x01e8);

	} else if (0x01f0 <= port && port < 0x01f8) {
		/* ide0 reg0-7 */
		value = umide_inb(0, port - 0x01f0);

	} else if (0x0280 <= port && port < 0x02a0) {
		/* NE2000 1 */
		value = umnet_inb(1, port - 0x0280);

	} else if (0x0300 <= port && port < 0x0320) {
		/* NE2000 0 */
		value = umnet_inb(0, port - 0x0300);	

	} else if (0x0320 <= port && port < 0x0340) {
		/* NE2000 2 */
		value = umnet_inb(2, port - 0x0320);	

	} else if (port == 0x0366) {
		/* ide5 reg8 */
		value = umide_inb(5, 8);

	} else if (port == 0x036e) {
		/* ide3 reg8 */
		value = umide_inb(3, 8);

	} else if (port == 0x0376) {
		/* ide1 reg8 */
		value = umide_inb(1, 8);

	} else if (0x0380 <= port && port < 0x03A0) {
		/* NE2000 3 */
		value = umnet_inb(3, port - 0x0380);	

	} else if (0x03c0 <= port && port < 0x03db) {	/* FIXME VOSSI */
		/* ega/vga */
		value = umvga_inb(port);

	} else if (port == 0x03e6) {
		/* ide4 reg8 */
		value = umide_inb(4, 8);

	} else if (port == 0x03ee) {
		/* ide2 reg8 */
		value = umide_inb(2, 8);

	} else if (port == 0x03f6) {
		/* ide0 reg8 */
		value = umide_inb(0, 8);

	} else {
		char buf[64];

		sprintf(buf, "reading byte from unassigned port 0x%04x\n",
				port);
		lotol_write(1, buf, strlen(buf));

		value = (unsigned char) in_res++;
	}

	return value;
}

void insb(unsigned short port, void *addr, unsigned long count)
{
	extern void umide_insb(int, unsigned short, void *, unsigned long);

	/* Keep list sorted by port number! */

	switch (port) {
	case 0x0160:	/* ide5 data reg */
		umide_insb(5, 0, addr, count);
		break;
	case 0x0168:	/* ide3 data reg */
		umide_insb(3, 0, addr, count);
		break;
	case 0x0170:	/* ide1 data reg */
		umide_insb(1, 0, addr, count);
		break;
	case 0x01e0:	/* ide4 data reg */
		umide_insb(4, 0, addr, count);
		break;
	case 0x01e8:	/* ide2 data reg */
		umide_insb(2, 0, addr, count);
		break;
	case 0x01f0:	/* ide0 data reg */
		umide_insb(0, 0, addr, count);
		break;
	default:
		/* FIXME */
		panic("WARNING: %s called with port=0x%04x\n",
				__FUNCTION__, port);
	}
}

unsigned short inw(unsigned short port)
{
	extern unsigned short umrtc_inw(unsigned short);
	unsigned short value;

	/* Keep list sorted by port nuber! */

	if (0x0020 <= port && port <= 0x0021) {
		/* pic (master) */
		assert(0);

	} else if (0x0040 <= port && port <= 0x0043) {
		/* pit0 */
		assert(0);

	} else if (0x0048 <= port && port <= 0x004b) {
		/* pit1 */
		assert(0);

	} if (0x0070 <= port && port <= 0x007f) {
		/* real time clock */
		value = umrtc_inw(port);

	} else if (0x00a0 <= port && port <= 0x00a1) {
		/* pic (slave) */
		assert(0);

	} else if (0x0060 == port || port == 0x0064) {
		/* keyboard */
		/*
		 * keyboard should do everthing bytewise so it should
		 * not be here.
		 */
		assert(0) ;

	/* ide missing FIXME VOSSI
	 * not needed anyway --tg 12:56 2002-06-29 */
		

	} else {
		char buf[64];

		sprintf(buf, "reading word from unassigned port 0x%04x\n",
				port);
		lotol_write(1, buf, strlen(buf));

		value = (unsigned short) in_res++;
	}

	return value;
}

void insw(unsigned short port, void *addr, unsigned long count)
{
	extern void umide_insw(int, unsigned short, void *, unsigned long);
	extern void umnet_insw(int, unsigned short, void *, unsigned long);

	/* Keep list sorted by port number! */

	switch (port) {
	case 0x0160:	/* ide5 data reg */
		umide_insw(5, 0x0160 - 0x0160, addr, count);
		break;
	case 0x0168:	/* ide3 data reg */
		umide_insw(3, 0x0168 - 0x0168, addr, count);
		break;
	case 0x0170:	/* ide1 data reg */
		umide_insw(1, 0x0170 - 0x0170, addr, count);
		break;
	case 0x01e0:	/* ide4 data reg */
		umide_insw(4, 0x01e0 - 0x01e0, addr, count);
		break;
	case 0x01e8:	/* ide2 data reg */
		umide_insw(2, 0x01e8 - 0x01e8, addr, count);
		break;
	case 0x01f0:	/* ide0 data reg */
		umide_insw(0, 0x01f0 - 0x01f0, addr, count);
		break;
	case 0x0290:     /* NE2000 1 data reg */
		umnet_insw(1, 0x0290 - 0x0280, addr, count);
		break;
	case 0x0310:     /* NE2000 0 data reg */
		umnet_insw(0, 0x0310 - 0x0300, addr, count);
		break;
	case 0x0330:     /* NE2000 2 data reg */
		umnet_insw(2, 0x0330 - 0x0320, addr, count);
		break;
	case 0x0390:     /* NE2000 3 data reg */
		umnet_insw(3, 0x0390 - 0x0380, addr, count);
		break;
	default:
		/* FIXME */
		panic("WARNING: %s called with port=0x%04x\n",
				__FUNCTION__, port);
	}
}

unsigned long inl(unsigned short port)
{
	unsigned long low;
	unsigned long high;

	low = inw(port);
	port += 2;
	high = inw(port);
	return (high << 16) | (low << 0);
}

void insl(unsigned short port, void *addr, unsigned long count)
{
	panic("%s called\n", __FUNCTION__);
}

void outb(unsigned char value, unsigned short port)
{
	/* Keep list sorted alphabetically! */
	extern void umapic_outb(unsigned char, unsigned short);
	extern void umide_outb(int, unsigned char, unsigned short);
	extern void umkbd_outb(unsigned char, unsigned short);
	extern void umnet_outb(int, unsigned char, unsigned short);
	extern void umpic_outb(int, unsigned char, unsigned short);
	extern void umpit_outb(int, unsigned char, unsigned short);
	extern void umppi_outb(unsigned char, unsigned short);
	extern void umrtc_outb(unsigned char, unsigned short);
	extern void umvga_outb(unsigned char, unsigned short);

	/* Keep list sorted by port number! */

	if (0x0020 <= port && port <= 0x0021) {
		/* pic (master) */
		umpic_outb(0, value, port - 0x0020);
	} else if (0x0022 <= port && port <= 0x0023) {
		/* IMCR (Interrupt Mode Configuration Register) */
		umapic_outb(value, port - 0x0022);

	} else if (0x0040 <= port && port <= 0x0043) {
		/* pit0 */
		umpit_outb(0, value, port - 0x0040);

	} else if (0x0048 <= port && port <= 0x004b) {
		/* pit1 */
		umpit_outb(1, value, port - 0x0048);

	} else if (0x0060 == port || port == 0x0064) {
		/* keyboard */
		umkbd_outb(value, port);

	} else if (0x0061 <= port && port <= 0x0063) {
		/* PPI 8255 (multi purpose I/O) */
		umppi_outb(value, port - 0x0061);

	} else if (0x0070 <= port && port <= 0x007f) {
		/* real time clock */
		umrtc_outb(value, port);

	} else if (port == 0x0080) {
		/* ????? */
		/*
		 * used to slow down i/o-operations
		 * (see include/asm-i386/io.h)
		 */
		/* do nothing... */

	} else if (0x00a0 <= port && port <= 0x00a1) {
		/* pic (slave) */
		umpic_outb(1, value, port - 0x00a0);

	} else if (0x0160 <= port && port < 0x0168) {
		/* ide5 reg0-7 */
		umide_outb(5, value, port - 0x0160);

	} else if (0x0168 <= port && port < 0x0170) {
		/* ide3 reg0-7 */
		umide_outb(3, value, port - 0x0168);

	} else if (0x0170 <= port && port < 0x0178) {
		/* ide1 reg0-7 */
		umide_outb(1, value, port - 0x0170);

	} else if (0x01e0 <= port && port < 0x01e8) {
		/* ide4 reg0-7 */
		umide_outb(4, value, port - 0x01e0);

	} else if (0x01e8 <= port && port < 0x01f0) {
		/* ide2 reg0-7 */
		umide_outb(2, value, port - 0x01e8);

	} else if (0x01f0 <= port && port < 0x01f8) {
		/* ide0 reg0-7 */
		umide_outb(0, value, port - 0x01f0);

	} else if (0x0280 <= port && port < 0x0300) {
		/* NE2000 1 */
		umnet_outb(1, value, port - 0x0280);

	} else if (0x0300 <= port && port < 0x0320) {
		/* NE2000 0 */
		umnet_outb(0, value, port - 0x0300);

	} else if (0x0320 <= port && port < 0x0340) {
		/* NE2000 2 */
		umnet_outb(2, value, port - 0x0320);

	} else if (port == 0x0366) {
		/* ide5 reg8 */
		umide_outb(5, value, 8);

	} else if (port == 0x036e) {
		/* ide3 reg8 */
		umide_outb(3, value, 8);

	} else if (port == 0x0376) {
		/* ide1 reg8 */
		umide_outb(1, value, 8);

	} else if (0x0380 <= port && port < 0x03a0) {
		/* NE2000 3 */
		umnet_outb(3, value, port - 0x0380);

	} else if (0x03c0 <= port && port < 0x03db) {	/* FIXME VOSSI */
		/* ega/vga */
		umvga_outb(value, port);

	} else if (port == 0x03e6) {
		/* ide4 reg8 */
		umide_outb(4, value, 8);

	} else if (port == 0x03ee) {
		/* ide2 reg8 */
		umide_outb(2, value, 8);

	} else if (port == 0x03f6) {
		/* ide0 reg8 */
		umide_outb(0, value, 8);

	} else {
		char buf[64];

		sprintf(buf, "writing byte 0x%02x to unassigned port 0x%04x\n",
				value, port);
		lotol_write(1, buf, strlen(buf));
	}
}

void outsb(unsigned short port, const void *addr, unsigned long count)
{
	extern void umide_outsb(int, unsigned short, const void *, unsigned long);

	/* Keep list sorted by port number! */

	switch (port) {
	case 0x0160:	/* ide5 data reg */
		umide_outsb(5, 0x0160 - 0x0160, addr, count);
		break;
	case 0x0168:	/* ide3 data reg */
		umide_outsb(3, 0x0168 - 0x0168, addr, count);
		break;
	case 0x0170:	/* ide1 data reg */
		umide_outsb(1, 0x0170 - 0x0170, addr, count);
		break;
	case 0x01e0:	/* ide4 data reg */
		umide_outsb(4, 0x01e0 - 0x01e0, addr, count);
		break;
	case 0x01e8:	/* ide2 data reg */
		umide_outsb(2, 0x01e8 - 0x01e8, addr, count);
		break;
	case 0x01f0:	/* ide0 data reg */
		umide_outsb(0, 0x01f0 - 0x01f0, addr, count);
		break;
	default:
		/* FIXME */
		panic("WARNING: %s called with port=0x%04x\n",
				__FUNCTION__, port);
	}
}

void outw(unsigned short value, unsigned short port)
{
	extern void umrtc_outw(unsigned short value, unsigned short port);

	/* Keep list sorted by port number! */

	if (0x0020 <= port && port <= 0x0021) {
		/* pic (master) */
		assert(0);

	} else if (0x0040 <= port && port <= 0x0043) {
		/* pit0 */
		assert(0);

	} else if (0x0048 <= port && port <= 0x004b) {
		/* pit1 */
		assert(0);

	} else if (0x0060 == port || port == 0x0064) {
		/* keyboard */
		/*
		 * keyboard should do everthing bytewise so it should
		 * not be here.
		 */
		assert(0) ;

	} else if (0x0070 <= port && port <= 0x007f) {
		/* real time clock */
		umrtc_outw(value, port);

	/* ide missing FIXME VOSSI */

	} else if (0x00a0 <= port && port <= 0x00a1) {
		/* pic (slave) */
		assert(0);

	} else {
		char buf[64];

		sprintf(buf, "writing word 0x%04x to unassigned port 0x%04x\n",
				value, port);
		lotol_write(1, buf, strlen(buf));
	}
}

void outsw(unsigned short port, const void *addr, unsigned long count)
{
	extern void umide_outsw(int, unsigned short, const void *, unsigned long);
	extern void umnet_outsw(int, unsigned short, const void *, unsigned long);

	/* Keep list sorted by port number! */

	switch (port) {
	case 0x0160:	/* ide5 data reg */
		umide_outsw(5, 0x0160 - 0x0160, addr, count);
		break;
	case 0x0168:	/* ide3 data reg */
		umide_outsw(3, 0x0168 - 0x0168, addr, count);
		break;
	case 0x0170:	/* ide1 data reg */
		umide_outsw(1, 0x0170 - 0x0170, addr, count);
		break;
	case 0x01e0:	/* ide4 data reg */
		umide_outsw(4, 0x01e0 - 0x01e0, addr, count);
		break;
	case 0x01e8:	/* ide2 data reg */
		umide_outsw(2, 0x01e8 - 0x01e8, addr, count);
		break;
	case 0x01f0:	/* ide0 data reg */
		umide_outsw(0, 0x01f0 - 0x01f0, addr, count);
		break;
	case 0x0290:     /* NE2000 1 data reg */
		umnet_outsw(1, 0x0290 - 0x0280, addr, count);
		break;
	case 0x0310:     /* NE2000 0 data reg */
		umnet_outsw(0, 0x0310 - 0x0300, addr, count);
		break;
	case 0x0330:     /* NE2000 2 data reg */
		umnet_outsw(2, 0x0330 - 0x0320, addr, count);
		break;
	case 0x0390:     /* NE2000 3 data reg */
		umnet_outsw(3, 0x0390 - 0x0380, addr, count);
		break;
	default:
		/* FIXME */
		panic("WARNING: %s called with port=0x%04x\n",
				__FUNCTION__, port);
	}
}

void outl(unsigned long value, unsigned short port)
{
	outw((unsigned short) (value & 0xffff), port);
	port += 2;
	outw((unsigned short) (value >> 16), port);
}

void outsl(unsigned short port, const void *addr, unsigned long count)
{
	panic("%s called\n", __FUNCTION__);
}
