#include <linux/assert.h>
#include <linux/sched.h>	/* needed by strncmp */
#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>

#define DEBUGGING 0

/* seek stuff */
# define SEEK_SET               0	/* Seek from beginning of file.  */
# define SEEK_CUR               1	/* Seek from current position.  */
# define SEEK_END               2	/* Seek from end of file.  */

/* FIXME: wrapper functions to avoid egcs compiler bug */
static int
lotol_read_wrapper (int fd, char *buf, int count)
{
	return lotol_read (fd, buf, count);
}

static int
lotol_write_wrapper (int fd, char *buf, int count)
{
	return lotol_write (fd, buf, count);
}

static int
lotol_lseek_wrapper (int fd, int offset, int whence)
{
	return lotol_lseek (fd, offset, whence);
}

#define LOTOL_OUT(fmt, arg...) \
do { \
	char buf[256]; \
	sprintf(buf, "%20s:% 4d: " fmt ,__FUNCTION__,__LINE__,##arg); \
	lotol_write_wrapper(1, buf, strlen(buf)); \
} while(0)

#if DEBUGGING
#define DEBUG(fmt, arg...) \
do { \
	char buf[256]; \
	sprintf(buf, "%20s:% 4d: eth%d: " fmt , \
			__FUNCTION__,__LINE__,nr,##arg); \
	lotol_write_wrapper(1, buf, strlen(buf)); \
} while(0)
#else
#define DEBUG(fmt, arg...) while(0) { }
#endif

#define ETHER_ADDR_LEN  6

/* Some generic ethernet register configurations. */
#define E8390_TX_IRQ_MASK       0xa     /* For register EN0_ISR */
#define E8390_RX_IRQ_MASK       0x5
#define E8390_RXCONFIG          0x4     /* EN0_RXCR: broadcasts, no multicast,
					   errors */
#define E8390_RXOFF             0x20    /* EN0_RXCR: Accept no packets */
#define E8390_TXCONFIG          0x00    /* EN0_TXCR: Normal transmit mode */
#define E8390_TXOFF             0x02    /* EN0_TXCR: Transmitter off */

#define PROMISCUOUS		0x10    /* HRM? What could that be? */

/*  Register accessed at EN_CMD, the 8390 base addr.  */
#define E8390_STOP      0x01    /* Stop and reset the chip */
#define E8390_START     0x02    /* Start the chip, clear reset */
#define E8390_TRANS     0x04    /* Transmit a frame */
#define E8390_RREAD     0x08    /* Remote read */
#define E8390_RWRITE    0x10    /* Remote write  */
#define E8390_NODMA     0x20    /* Remote DMA */
#define E8390_PAGE0     0x00    /* Select page chip registers */
#define E8390_PAGE1     0x40    /* using the two high-order bits */
#define E8390_PAGE2     0x80    /* Page 3 is invalid. */

/* --tg 9:06 2002-06-26 INTERRUPT REGISTER */
#define PRXE		0x01	/* Packet Received Interrupt Enable */
#define PTXE		0x02	/* Packet Transmit Interrupt Enable */
#define PXEE		0x04	/* Packet Reiceive Error Interrupt Enable */
#define TXEE		0x08	/* Packet Transmit Error Interrupt Enable */
#define OVWE		0x10	/* Overwrite Warning Interrupt Enable */
#define CNTE		0x20	/* Counter Overflow Interrupt Enable */
#define RDCE		0x40	/* DMA Complete Interrupt Enable */

#define NE_BASE  (dev->base_addr)
#define NE_CMD          0x00
#define NE_DATAPORT     0x10    /* NatSemi-defined port window offset. */
#define NE_RESET        0x1f    /* Issue a read to reset, a write to clear. */
#define NE_IO_EXTENT    0x20

#define NESM_START_PG   0x40    /* First page of TX buffer */
#define NESM_STOP_PG    0x80    /* Last page +1 of RX ring */

#define E8390_CMD       0x00  /* The command register (for all pages) */
/* Page 0 register offsets. */
#define EN0_CLDALO      0x01  /* Low byte of current local dma addr  RD */
#define EN0_STARTPG     0x01  /* Starting page of ring bfr WR */
#define EN0_CLDAHI      0x02  /* High byte of current local dma addr  RD */  
#define EN0_STOPPG      0x02  /* Ending page +1 of ring bfr WR */
#define EN0_BOUNDARY    0x03  /* Boundary page of ring bfr RD WR */
#define EN0_TSR         0x04  /* Transmit status reg RD */
#define EN0_TPSR        0x04  /* Transmit starting page WR */
#define EN0_NCR         0x05  /* Number of collision reg RD */
#define EN0_TCNTLO      0x05  /* Low  byte of tx byte count WR */
#define EN0_FIFO        0x06  /* FIFO RD */
#define EN0_TCNTHI      0x06  /* High byte of tx byte count WR */
#define EN0_ISR         0x07  /* Interrupt status reg RD WR */
#define EN0_CRDALO      0x08  /* low byte of current remote dma address RD */
#define EN0_RSARLO      0x08  /* Remote start address reg 0 */
#define EN0_CRDAHI      0x09  /* high byte, current remote dma address RD */ 
#define EN0_RSARHI      0x09  /* Remote start address reg 1 */
#define EN0_RCNTLO      0x0a  /* Remote byte count reg WR */
#define EN0_RCNTHI      0x0b  /* Remote byte count reg WR */
#define EN0_RSR         0x0c  /* rx status reg RD */
#define EN0_RXCR        0x0c  /* RX configuration reg WR */
#define EN0_TXCR        0x0d  /* TX configuration reg WR */
#define EN0_COUNTER0    0x0d  /* Rcv alignment error counter RD */
#define EN0_DCFG        0x0e  /* Data configuration reg WR */
#define EN0_COUNTER1    0x0e  /* Rcv CRC error counter RD */
#define EN0_IMR         0x0f  /* Interrupt mask reg WR */
#define EN0_COUNTER2    0x0f  /* Rcv missed frame error counter RD */

/* Bits in EN0_ISR - Interrupt status register */
#define ENISR_RX        0x01    /* Receiver, no error */
#define ENISR_TX        0x02    /* Transmitter, no error */
#define ENISR_RX_ERR    0x04    /* Receiver, with error */
#define ENISR_TX_ERR    0x08    /* Transmitter, with error */
#define ENISR_OVER      0x10    /* Receiver overwrote the ring */
#define ENISR_COUNTERS  0x20    /* Counters need emptying */
#define ENISR_RDC       0x40    /* remote dma complete */
#define ENISR_RESET     0x80    /* Reset completed */
#define ENISR_ALL       0x3f    /* Interrupts we will enable */

/* Bits in received packet status byte and EN0_RSR*/
#define ENRSR_RXOK      0x01    /* Received a good packet */
#define ENRSR_CRC       0x02    /* CRC error */
#define ENRSR_FAE       0x04    /* frame alignment error */
#define ENRSR_FO        0x08    /* FIFO overrun */
#define ENRSR_MPA       0x10    /* missed pkt */
#define ENRSR_PHY       0x20    /* physical/multicast address */
#define ENRSR_DIS       0x40    /* receiver disable. set in monitor mode */
#define ENRSR_DEF       0x80    /* deferring */

/* Transmitted packet status, EN0_TSR. */
#define ENTSR_PTX 0x01  /* Packet transmitted without error */
#define ENTSR_ND  0x02  /* The transmit wasn't deferred. */
#define ENTSR_COL 0x04  /* The transmit collided at least once. */
#define ENTSR_ABT 0x08  /* The transmit collided 16 times, and was deferred. */
#define ENTSR_CRS 0x10  /* The carrier sense was lost. */
#define ENTSR_FU  0x20  /* A "FIFO underrun" occurred during transmit. */
#define ENTSR_CDH 0x40  /* The collision detect "heartbeat" signal was lost. */
#define ENTSR_OWC 0x80  /* There was an out-of-window collision. */

/* --tg 17:02 2002-06-23 */
#define EN1_MAC0       0x01    /* This board's physical enet addr RD WR */
#define EN1_MAC1       0x02    /* This board's physical enet addr RD WR */
#define EN1_MAC2       0x03    /* This board's physical enet addr RD WR */
#define EN1_MAC3       0x04    /* This board's physical enet addr RD WR */
#define EN1_MAC4       0x05    /* This board's physical enet addr RD WR */
#define EN1_MAC5       0x06    /* This board's physical enet addr RD WR */
#define EN1_CURPAG     0x07    /* Current memory page RD WR */
#define EN1_MULT0      0x08    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT1      0x09    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT2      0x0a    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT3      0x0b    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT4      0x0c    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT5      0x0d    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT6      0x0e    /* Multicast filter mask array (8 bytes) RD WR */
#define EN1_MULT7      0x0f    /* Multicast filter mask array (8 bytes) RD WR */

struct umnet {
	int myself; /* Has to be initalized to -1 if card not used */
	int fd;

	struct {
		unsigned char mac[ETHER_ADDR_LEN];
		long addr;
		int port;
	} host[16];

	
	/* FIXME could be one buffer */
	unsigned char prom[32];
	unsigned char rbuf[13312]; /* ring buffer */ /* PAY MORE ATTENTION THOMAS! */
	unsigned char tbuf[4096];  /* send buffer */

	unsigned int  count;
	unsigned int  start;    /*  0x08   Remote start address */

	unsigned char cmd;      /*  0x00   The command register (for all pages) */
	
	/* PAGE0 */
	unsigned char boundary; /*  0x03   Boundary page of ring bfr RD WR */
	unsigned char tsr;      /*  0x04   Transmit status reg RD */

	unsigned int  tpsr;     /*  0x04   Transmit starting page WR */

	unsigned char isr;      /*  0x07   Interrupt status reg RD WR */
	unsigned char rsr;      /*  0x0c   rx status reg RD */
	unsigned char rxcr;     /*  0x0c   RX configuration reg WR */
	unsigned char txcr;     /*  0x0d   TX configuration reg WR */
	unsigned char counter0; /*  0x0d   Rcv alignment error counter RD */
	unsigned char dcfg;     /*  0x0e   Data configuration reg WR */
	unsigned char counter1; /*  0x0e   Rcv CRC error counter RD */
	unsigned char imr;      /*  0x0f   Interrupt mask reg WR */
	unsigned char counter2; /*  0x0f   Rcv missed frame error counter RD */

	/* PAGE1 */
	unsigned char mac0;    /* 0x01 hw addr */ 
	unsigned char mac1;    /* 0x02 */
	unsigned char mac2;    /* 0x03 */
	unsigned char mac3;    /* 0x04 */
	unsigned char mac4;    /* 0x05 */
	unsigned char mac5;    /* 0x06 */

	unsigned char curpag;  /* 0x07 Current Memmory Page*/

	unsigned char mult0;   /* 0x08 Multicast Array */
	unsigned char mult1;   /* 0x09 */
	unsigned char mult2;   /* 0x0a */
	unsigned char mult3;   /* 0x0b */
	unsigned char mult4;   /* 0x0c */
	unsigned char mult5;   /* 0x0d */
	unsigned char mult6;   /* 0x0e */
	unsigned char mult7;   /* 0x0f */
} card[4] = { /* card[0-4].myself = -1;  see comment in umide.h side effect */
	{ -1 },
	{ -1 },
	{ -1 },
	{ -1 },
};

int umnet_irqs[4] = { 3, 4, 6, 7 }; /* FIXME make configurable */

struct umnet_e8390_pkt_hdr {
	unsigned char  status;
	unsigned char  next;
	unsigned short count;
};

#if DEBUGGING
/* debugging purposes */
static struct { unsigned char txt[80]; } umnet_page0_out[] = {
	{ "E8390_CMD    The command register (for all pages)" },      /* 0x00 */
	{ "EN0_STARTPG  Starting page of ring bfr WR" },              /* 0x01 */
	{ "EN0_STOPPG   Ending page +1 of ring bfr WR" },             /* 0x02 */
	{ "EN0_BOUNDARY Boundary page of ring bfr RD WR" },           /* 0x03 */
	{ "EN0_TSR      Transmit status reg RD" },                    /* 0x04 */
	{ "EN0_TCNTLO   Low  byte of tx byte count WR" },             /* 0x05 */
	{ "EN0_TCNTHI   High byte of tx byte count WR" },             /* 0x06 */
	{ "EN0_ISR      Interrupt status reg RD WR" },                /* 0x07 */
	{ "EN0_RSARLO   Remote start address reg 0" },                /* 0x08 */
	{ "EN0_RSARHI   Remote start address reg 1" },                /* 0x09 */
	{ "EN0_RCNTLO   Remote byte count reg WR" },                  /* 0x0a */
	{ "EN0_RCNTHI   Remote byte count reg WR" },                  /* 0x0b */
	{ "EN0_RXCR     RX configuration reg WR" },                   /* 0x0c */
	{ "EN0_TXCR     TX configuration reg WR" },                   /* 0x0d */
	{ "EN0_DCFG     Data configuration reg WR" },                 /* 0x0e */
	{ "EN0_IMR      Interrupt mask reg WR" }                      /* 0x0f */
};

static struct { unsigned char txt[80]; } umnet_page0_in[] = {
	{ "E8390_CMD    The command register (for all pages)" },      /* 0x00 */
	{ "EN0_CLDALO   Low byte of current local dma addr  RD" },    /* 0x01 */
 	{ "EN0_CLDAHI   High byte of current local dma addr  RD" },   /* 0x02 */
	{ "EN0_BOUNDARY Boundary page of ring bfr RD WR" },           /* 0x03 */
 	{ "EN0_TSR      Transmit status reg RD" },                    /* 0x04 */
 	{ "EN0_NCR      Number of collision reg RD" },                /* 0x05 */
 	{ "EN0_FIFO     FIFO RD" },                                   /* 0x06 */
	{ "EN0_ISR      Interrupt status reg RD WR" },                /* 0x07 */
 	{ "EN0_CRDALO   low byte of current remote dma address RD" }, /* 0x08 */
 	{ "EN0_CRDAHI   high byte, current remote dma address RD" },  /* 0x09 */
	{ "Please fill in" },
	{ "Please fill in" },
 	{ "EN0_RSR      rx status reg RD" },                          /* 0x0c */
 	{ "EN0_COUNTER0 Rcv alignment error counter RD" },            /* 0x0d */
 	{ "EN0_COUNTER1 Rcv CRC error counter RD" },                  /* 0x0e */
 	{ "EN0_COUNTER2 Rcv missed frame error counter RD" }          /* 0x0f */
};

static struct { unsigned char txt[80]; } umnet_page1[] = {
	{ "E8390_CMD    The command register (for all pages)" },      /* 0x00 */
	{ "EN1_MAC0     This board's physical enet addr RD WR" },     /* 0x01 */
	{ "EN1_MAC1     This board's physical enet addr RD WR" },     /* 0x02 */
	{ "EN1_MAC2     This board's physical enet addr RD WR" },     /* 0x03 */
	{ "EN1_MAC3     This board's physical enet addr RD WR" },     /* 0x04 */
	{ "EN1_MAC4     This board's physical enet addr RD WR" },     /* 0x05 */
	{ "EN1_MAC5     This board's physical enet addr RD WR" },     /* 0x06 */
	{ "EN1_CURPAG   Current memory page RD WR" },                 /* 0x07 */
	{ "EN1_MULT0    Multicast filter mask array (8 bytes) RD WR"},/* 0x08 */
	{ "EN1_MULT1    Multicast filter mask array (8 bytes) RD WR"},/* 0x09 */
	{ "EN1_MULT2    Multicast filter mask array (8 bytes) RD WR"},/* 0x0a */
	{ "EN1_MULT3    Multicast filter mask array (8 bytes) RD WR"},/* 0x0b */
	{ "EN1_MULT4    Multicast filter mask array (8 bytes) RD WR"},/* 0x0c */
	{ "EN1_MULT5    Multicast filter mask array (8 bytes) RD WR"},/* 0x0d */
	{ "EN1_MULT6    Multicast filter mask array (8 bytes) RD WR"},/* 0x0e */
	{ "EN1_MULT7    Multicast filter mask array (8 bytes) RD WR"} /* 0x0f */
};
#endif
