/*
 * $Id: umpic.c,v 1.1.1.1 2002/08/23 13:52:08 kingst Exp $
 */

#include <linux/assert.h>
#include <asm/lotol.h>


static unsigned char step[2] = { 0, 0 };
static unsigned char imr[2] = { 0xff, 0xff };
static unsigned char irr[2] = { 0x00, 0x00 };
static unsigned char isr[2] = { 0x00, 0x00 };
static unsigned char vec[2] = { 0x20, 0x28 };
static unsigned char sngl[2] = { 0, 0 };
static unsigned char ic4[2] = { 0, 0 };
static unsigned char ris[2] = { 0, 0 };


int
umpic_test_irq(void)
{
	if (isr[0] != 0 || isr[1] != 0) {
		return 0;
	}
	if ((irr[0] & ~imr[0])
	 || (irr[1] & ~imr[1])) {
		return 1;
	} else {
		return 0;
	}
}

unsigned char
umpic_ack_irq(void)
{
	int slave;
	int irq;

	for (slave = 0; slave < 2; slave++) {
		unsigned char m;
		m = irr[slave] & ~imr[slave];
		for (irq = 0; irq < 8; irq++) {
			if (m & (1 << irq)) {
				isr[slave] |= 1 << irq;
				irr[slave] &= ~(1 << irq);
				return vec[slave] + irq;
			}
		}
	}

	assert(0);
	return -1;
}

void
call_isa_irq(int irq)
{
	extern void umcore_scan_irqs(void);
#ifdef CONFIG_SMP
	extern int umapic_test_lint(int);
#endif	/* CONFIG_SMP */

	assert(irq != 2);

	if (irq < 8) {
		irr[0] |= 1 << irq;
	} else {
		irr[1] |= 1 << (irq - 8);
		if (irr[1] & ~imr[1]) {
			irr[0] |= 1 << 2;
		}
	}

#ifdef CONFIG_SMP
	/*
	 * See, if this interrupt is attached to the local CPU interrupt pins.  
	 * If yes, check wether this one is masked or not.
	 */
	if (umapic_test_lint(irq)) {
		return;
	}
#endif	/* CONFIG_SMP */

	umcore_scan_irqs();
}

unsigned char
umpic_inb(int slave, unsigned short port)
{
	unsigned char value;

	switch (port) {
	case 0: /* 0x20 / 0xa0 */
		if (ris[slave]) {
			value = isr[slave];
		} else {
			value = irr[slave];
		}
		break;

	case 1: /* 0x21 / 0xa1 */
		value = imr[slave];
		break;

	default:
		value = 0;
		assert(0);
	}

	return value;
}

void
umpic_outb(int slave, unsigned char value, unsigned short port)
{
	extern void umcore_scan_irqs(void);
	int irq;

	switch (port) {
	case 0: /* 0x20 / 0xa0 */
		if (value & 0x10) {
			/* ICW1 */
			/* ltim[slave] = (value >> 3) & 1; */
			sngl[slave] = (value >> 1) & 1;
			ic4[slave] = (value >> 0) & 1;
			step[slave] = 1;
			break;

		} else {
			if (! (value & 0x08)) {
				/* OCW2 */
				switch (value & 0xe0) {
				case 0x00:
					/* FIXME VOSSI */
					assert(0);
					break;

				case 0x20:
					/* unspecific EOI */
					for (irq = 0; ; irq++) {
						assert(irq < 8);
						if (isr[slave] & (1 << irq)) {
							break;
						}
					}
					isr[slave] &= ~(1 << irq);
					umcore_scan_irqs();
					break;

				case 0x40:
					/* FIXME VOSSI */
					assert(0);
					break;

				case 0x60:
					/* specific EOI */
					isr[slave] &= ~(1 << (value & 7));
					umcore_scan_irqs();
					break;

				case 0x80:
					/* FIXME VOSSI */
					assert(0);
					break;

				case 0xa0:
					/* FIXME VOSSI */
					assert(0);
					break;

				case 0xc0:
					/* FIXME VOSSI */
					assert(0);
					break;

				case 0xe0:
					/* FIXME VOSSI */
					assert(0);
					break;
				}
			} else {
				/* OCW3 */
				/* FIXME VOSSI */
				if (value & 2) {
					ris[slave] = (value >> 0) & 1;
				}
			}
		}
		break;

	case 1: /* 0x21 / 0xa1 */
		switch (step[slave]) {
		case 0:
			/* OCW1 */
			imr[slave] = value;

			umcore_scan_irqs();
			break;

		case 1:
			/* ICW2 */
			assert((value & 0x07) == 0);
			assert((slave == 0 && value == 0x20)
			    || (slave == 1 && value == 0x28));

			vec[slave] = value;

			if (sngl[slave] == 0) {
				step[slave] = 2;
			} else if (ic4[slave]) {
				step[slave] = 3;
			} else {
				step[slave] = 0;
			}
			break;

		case 2:
			/* ICW3 */
			/* ignored */
			assert((slave == 0 && value == (1 << 2))
			    || (slave == 1 && value == 2));

			if (ic4[slave]) {
				step[slave] = 3;
			} else {
				step[slave] = 0;
			}
			break;

		case 3:
			/* ICW4 */
			/* ignored */
			step[slave] = 0;
			break;

		default:
			assert(0);
		}
		break;

	default:
		assert(0);
	}
}

void
umpic_init(void)
{
}
