/*
 * $Id: umpit.c,v 1.2 2002/10/04 14:49:08 dunlapg Exp $
 */

#include <linux/assert.h>
#include <asm/lotol.h>

#define UMPIT_REAL_HZ	1193180
#define UMPIT_EMUL_HZ	20

/* calibrate 10ms to tsc */
#define CALIBRATE_TIME	(1000000/UMPIT_EMUL_HZ)

/* read per-cpu tsc */
extern void umcpu_tsc_get(unsigned long *, unsigned long *);

static unsigned long last_tsc_low = 0;	/* lsb 32 bits of Time Stamp Counter */ 
static unsigned long last_tsc_high = 0;	/* msb 32 bits of Time Stamp Counter */ 
static unsigned long tsc_to_timer = 0; /* ticks per timer interval */

static struct {
	unsigned char cr;
	unsigned char modus[3];
	unsigned short start[3];
	unsigned short act[3];		/* only used for one-shot timer */
	unsigned short latch[3];
	unsigned int gate[3];
} umpit[2] = {
	{ -1, /* ... */ },
	{ -1, /* ... */ },
};

/* 
 * 64-bit subtract - gcc just messes up with long longs.  Substract val1 from
 * val0.  
 *
 * XXX This belongs into umlib.c, as it is also needed somewhere else!
 */
void
sub64(unsigned long *val0_low, unsigned long *val0_high,
      unsigned long *val1_low, unsigned long *val1_high)
{
	asm ("subl %2,%0\n\t"
	     "sbbl %3,%1"
	     :"=a" (*val0_low), "=d" (*val0_high)
	     :"g" (*val1_low), "g" (*val1_high),
	      "0" (*val0_low), "1" (*val0_high));     
}

void
umpit_set_gate(int nr, int gate)
{
	umpit[nr].gate[gate] = 1;
}

void
umpit_clear_gate(int nr, int gate)
{
	umpit[nr].gate[gate] = 0;
}

int
umpit_get_timer(int nr, int port)
{
	unsigned long cur_tsc_low;
	unsigned long cur_tsc_high;
	unsigned long delta;
	int val;

	umcpu_tsc_get(&cur_tsc_low, &cur_tsc_high);

	sub64(&cur_tsc_low, &cur_tsc_high, &last_tsc_low, &last_tsc_high);

	delta = cur_tsc_low / tsc_to_timer;
	val = umpit[nr].start[port] - delta;

	return val;
}

void
umpit_interrupt(void)
{
	extern void call_isa_irq(int irq);

	int nr;
	int ch;

	for (nr = 0; nr < 2; nr++) {
		for (ch = 0; ch < 3; ch++) {
			switch (umpit[nr].modus[ch]) {
			case 0x00:	/* modus 0 binary */
				/* generate always an irq for clock timer */
				if (nr == 0 && ch == 0) {
					call_isa_irq(0);
				}
			
				break;

			case 0x04:	/* modus 2 binary */
				if (nr == 0 && ch == 0) {
					call_isa_irq(0);
				}

			case 0x06:	/* modus 3 binary */
				/* used for "beep" */
				/* FIXME VOSSI */
				break;

			default:
				{
					char buf[256];
					sprintf(buf, "umpit[%d].modus[%d]=%d\n",
						nr, ch, umpit[nr].modus[ch]);
					lotol_write(1, buf, strlen(buf));
				}
				assert(0);
			}
		}
	}
}

unsigned char
umpit_inb(int nr, unsigned short port)
{
	unsigned char value;

	value = 0;	/* FIXME VOSSI */

	switch (nr) {
	case 0:
		switch (port) {
		case 0: /* clock timer */
		case 1: /* memory refresh timer */
		case 2: /* speaker timer */
			assert(port == 0 || port == 2);

			if (umpit[nr].cr & 0x10) {
				/* should read low byte of latch */
				value = (umpit[nr].latch[port] >> 0) & 0xff;
				umpit[nr].cr &= ~0x10;

			} else if (umpit[nr].cr & 0x20) {
				/* should read high byte of latch */
				value = (umpit[nr].latch[port] >> 8) & 0xff;
				umpit[nr].cr &= ~0x20;

			} else {
				assert(0);
			}
			break;

		case 3:
			/* control register */
			/* write-only */
			assert(0);

		default:
			assert(0);
		}
		break;

	case 1:
		switch (port) {
		case 0:
			/* fail-safe timer */
			/* ignored */
			break;

		case 1:
			/* unused */
			/* ignored */
			break;

		case 2:
			/* unused */
			/* ignored */
			break;

		case 3:
			/* control register */
			/* write-only */
			assert(0);

		default:
			assert(0);
		}
		break;

	default:
		assert(0);
	}

	return value;
}

void
umpit_outb(int nr, unsigned char value, unsigned short port)
{
	unsigned long cur_tsc_low;
	unsigned long cur_tsc_high;
	unsigned long delta;

	extern void umppi_clear_nme(void);

	switch (nr) {
	case 0:
		switch (port) {
		case 0: /* clock timer */
		case 1: /* memory refresh timer */
		case 2: /* speaker timer */
			assert(port == 0 || port == 2);

			if (umpit[nr].cr & 0x10) {
				/* should write low byte of counter */
				umpit[nr].start[port] &= 0xff00;
				umpit[nr].start[port] |= value << 0;
				umpit[nr].cr &= ~0x10;

			} else if (umpit[nr].cr & 0x20) {
				/* should write high byte of counter */
				umpit[nr].start[port] &= 0x00ff;
				umpit[nr].start[port] |= value << 8;
				umpit[nr].cr &= ~0x20;
			}

			if ((umpit[nr].cr & 0x30) == 0) {
				/* start timer */
				umpit[nr].modus[port] = umpit[nr].cr & 0x0f;

				/* only for speaker timer */
				if (port == 2) {
					umppi_clear_nme();
					umpit[nr].act[port] 
						= umpit[nr].start[port];
				}

				/* remember current tsc */
				umcpu_tsc_get(&last_tsc_low, &last_tsc_high);
			}
			break;

		case 3:
			/* control register */
			assert((value >> 6) <= 2);
				/* we have no read-back feature */

			if (((value >> 4) & 0x03) == 0) {
				/* latch low/high byte of counter */

				value |= 0x03 << 4;

				umcpu_tsc_get(&cur_tsc_low, &cur_tsc_high);

				sub64(&cur_tsc_low, &cur_tsc_high, 
						&last_tsc_low, &last_tsc_high);

				delta = cur_tsc_low / tsc_to_timer;
				delta = delta % umpit[nr].start[value >> 6];
				umpit[nr].latch[value >> 6]
					= umpit[nr].start[value >> 6] - delta;
			}
			umpit[nr].cr = value;
			break;

		default:
			assert(0);
		}
		break;

	case 1:
		switch (port) {
		case 0:
			/* fail-safe timer */
			/* ignored */
			break;

		case 1:
			/* unused */
			/* ignored */
			break;

		case 2:
			/* unused */
			/* ignored */
			break;

		case 3:
			/* control register */
			/* ignored */
			break;

		default:
			assert(0);
		}
		break;

	default:
		assert(0);
	}
}

void
umpit_calibrate(void)
{
	struct lotol_timeval tvstart, tvend, tvtmp;
	unsigned long startlow, starthigh;
	unsigned long endlow, endhigh;
	int ret;

	ret = lotol_gettimeofday(&tvtmp, NULL);
	tvstart.tv_sec = tvtmp.tv_sec;
	tvstart.tv_usec = tvtmp.tv_usec;
	tvend.tv_sec = tvtmp.tv_sec + CALIBRATE_TIME / 1000000;
	tvend.tv_usec = tvtmp.tv_usec + CALIBRATE_TIME % 1000000;

	/* normalize microseconds */
	while (1000000 <= tvend.tv_usec) {
		tvend.tv_sec++;
		tvend.tv_usec -= 1000000;
	}

	/* calibrate */
	umcpu_tsc_get(&startlow, &starthigh);
	do {
		ret = lotol_gettimeofday(&tvtmp, NULL);
		assert(0 <= ret);
	} while (tvtmp.tv_sec < tvend.tv_sec || tvtmp.tv_usec < tvend.tv_usec);
	umcpu_tsc_get(&endlow, &endhigh);

	sub64(&endlow, &endhigh, &startlow, &starthigh);
	assert(endhigh == 0);

	/* tsc ticks per timer ticks */
	tsc_to_timer = endlow / (UMPIT_REAL_HZ / UMPIT_EMUL_HZ);

	return;
}

void
umpit_init(void)
{
	struct lotol_itimerval it;
	int ret;

	umpit_calibrate();

	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 1000000 / UMPIT_EMUL_HZ;
	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = 1000000 / UMPIT_EMUL_HZ;

	ret = lotol_setitimer(LOTOL_ITIMER_REAL, &it, NULL);
	assert(0 <= ret);
}
