#include <linux/config.h>
#include <linux/assert.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/pgtable.h>

#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>

//#define PAGE_OFFSET_RAW 0x70000000
#define PAGE_OFFSET_RAW __PAGE_OFFSET

#define CL_MAGIC_ADDR	((unsigned short *) (PAGE_OFFSET_RAW + 0x90020))
#define CL_MAGIC	0xA33F
#define CL_BASE_ADDR	((char *) PAGE_OFFSET_RAW + 0x90000)
#define CL_OFFSET	((unsigned short *) (PAGE_OFFSET_RAW + 0x90022))
#define COMMAND_LINE    ((char *) (PAGE_OFFSET_RAW+2048))
#define COMMAND_LINE_SIZE 256

int ap = 0;		/* default: 0 BSP (or uniprocessor) */

struct {
	void *esp;
	unsigned short ss;
} stack_start = { &init_stack[sizeof(init_stack) / sizeof(init_stack[0])], 0 };

void
_stext(int id)	/* some use _stext, some use stext... */
{
	void stext(int);

	stext(id);
}

void
stext(int id)
{
	extern void umcore_init(void);
	extern void umpic_init(void);
	extern void umpit_init(void);
	extern void umppi_init(void);
	extern void umeth_init(void);
	extern void umide_init(void);
	extern void umkbd_init(void);
	extern void umrtc_init(void);
	extern void umvga_init(void);
	extern void umapic_init(void);
	static void stext1(void);
	static void stext2(void);

	ap = id;

	/* This needs to be done by every CPU */
	umapic_init();
	umcore_init();

	/* This only by the BSP */
	if (!ap) {
		umpic_init();
		umpit_init();
		umppi_init();
		umeth_init();
		umide_init();
		umkbd_init();
		umrtc_init();
		umvga_init();
	}

	stext1();

	/* Set up the stack pointer */
	assert(sizeof(init_stack) == 2 * PAGE_SIZE);
	asm volatile(
		"movl %0, %%esp\n\t"
		: /* no output */
		: "p" (stack_start.esp)
	);

	stext2();
}

static void
stext1(void)
{
	int ret, i;
       	unsigned long p;
	
	if (!ap) {
		/*
		 * This part belongs to arch/i386/boot/...
		 */
		/* getting system data from the BIOS (linux host) */
		ret = lotol_lseek(FD_MAIN_MEMORY, 0, 2);
		assert(0 <= ret);
		ret -= 0x100000;	/* subtract "low" memory */
		ret /= 1024;		/* get number of kbytes */
		*(unsigned short *) (CL_BASE_ADDR + 0x002) = ret;	/* EXT_MEM_K */
		*(unsigned long *) (CL_BASE_ADDR + 0x1e0) = ret;	/* ALT_MEM_K */

		/* We have enough memory, setup e820 maps */
		if (0x100000 > ret) {
			unsigned long *e820;

#ifdef CONFIG_SMP
			*(unsigned short *)(CL_BASE_ADDR + 0x1e8) = 6;    /* E820_MAP_NR */
#else
			*(unsigned short *)(CL_BASE_ADDR + 0x1e8) = 3;    /* E820_MAP_NR */
#endif	/* CONFIG_SMP */
			e820 = ((unsigned long *)(CL_BASE_ADDR + 0x2d0)); /* E820_MAP */
			
			/* Base memory 0 to 640K */
			e820[0] = 0x00000000;
			e820[2] = 0x000A0000;
			e820[4] = 0x0001;	

			/* BIOS/VideoRAM/etc. 640K to 1M */
			e820[5] = 0x000A0000;
			e820[7] = 0x00060000;
			e820[9] = 0x0002;	

			/* Extended Memory 1M to MEMSIZE M */
			e820[10] = 0x00100000;
			e820[12] = ret * 1024;
			e820[14] = 0x0001;	

#ifdef CONFIG_SMP
			/* 4K I/O APIC */
			e820[15] = 0xFEC00000;
			e820[17] = 0x1000;
			e820[19] = 0x0002;	

			/* 4K local APIC */
			e820[20] = 0xFEE00000;
			e820[22] = 0x1000;
			e820[24] = 0x0002;	

			/* Fixed mappings */
			e820[25] = 0x7FFF0000;
			e820[27] = 0x10000;
			e820[29] = 0x0002;	
#endif /* CONFIG_SMP */
		}
	}

	/*
	 * Initialize page tables
	 */
	for (i = 0, p = 0x000007; i < 1024; i++, p += 0x1000) {
		pg0[i] = p;
	}
	for (i = 0, p = 0x400007; i < 1024; i++, p += 0x1000) {
		extern unsigned long pg1[1024];	/* missing in pgtable.h */
		pg1[i] = p;
	}
	pgd_val(swapper_pg_dir[0                >> PGDIR_SHIFT]) = 0x00102007;
	pgd_val(swapper_pg_dir[(0               >> PGDIR_SHIFT)+1]) =0x00103007;
	pgd_val(swapper_pg_dir[PAGE_OFFSET_RAW  >> PGDIR_SHIFT]) = 0x00102007;
	pgd_val(swapper_pg_dir[(PAGE_OFFSET_RAW >> PGDIR_SHIFT)+1]) =0x00103007;
	{
		extern void umcpu_cr3_set(unsigned long);
		umcpu_cr3_set(0x101000);
	}

	/* disable interrupts */
	cli();
}

static void
stext2(void)
{
	static void check_x87(void);
	extern void start_kernel(void);
#ifdef CONFIG_SMP
	extern void start_secondary(void);
#endif	/* CONFIG_SMP */
	int ret;

	/*
	 * Clear BSS first so that there are no surprises...
	 */
	if (!ap) {
		extern char __bss_start;
		extern char _end;
		char *p;

		for (p = &__bss_start; p < &_end; p++)
			*p = 0;
	}

	/* ... */

	/*
	 * Initialize eflags.  Some BIOS's leave bits like NT set.  This would
	 * confuse the debugger if this code is traced.
	 * XXX - best to initialize before switching to protected mode.
	 */
	asm volatile(
		"pushl $0\n\t"
		"popfl\n\t"
	);

	/*
	 * Copy bootup parameters out of the way. First 2kB of
	 * _empty_zero_page is for boot parameters, second 2kB
	 * is for the command line.
	 */
	if (!ap) {
		memcpy(empty_zero_page, (char *) CL_BASE_ADDR, 2048);

		if (*CL_MAGIC_ADDR == CL_MAGIC) {
			memcpy((char *) empty_zero_page + 2048,
				(char *) CL_BASE_ADDR + *CL_OFFSET, 2048);
		}
	}

	boot_cpu_data.cpuid_level = -1;		/* -1 for no CPUID initially */

	/* check if it is 486 or 386. */
	boot_cpu_data.x86 = 3;			/* at least 386 */
	
	asm volatile(
		"pushfl\n\t"
		"popl %%eax\n\t"
		"movl %%eax, %%ecx\n\t"
		"xorl $0x40000, %%eax\n\t"
		"pushl %%eax\n\t"
		"popfl\n\t"
		"pushfl\n\t"
		"popl %%eax\n\t"
		"xorl %%ecx, %%eax\n\t"
		"andl $0x40000, %%eax\n\t"
		"je is386\n\t"
		"movl $0, %0\n\t"
		"jmp 1f\n\t"
		"is386:\n\t"
		"movl $1, %0\n\t"
		"1:\n\t"
		"pushl %%ecx\n\t"		/* restore flags */
		"popfl\n\t"
	 	: "=r" (ret)
		: /* no input */
		: "eax" ,"ecx"
	);
	if (ret)
		goto is386;

	boot_cpu_data.x86 = 4;			/* at least 486 */
	asm volatile(
		"pushfl\n\t"
		"popl %%eax\n\t"
		"movl %%eax, %%ecx\n\t"
		"xorl $0x200000, %%eax\n\t"
		"pushl %%eax\n\t"
		"popfl\n\t"
		"pushfl\n\t"
		"popl %%eax\n\t"
		"xorl %%ecx, %%eax\n\t"
		"andl $0x200000, %%eax\n\t"
		"je is486\n\t"
		"movl $0, %0\n\t"
		"jmp 1f\n\t"
		"is486:\n\t"
		"movl $1, %0\n\t"
		"1:\n\t"
		: "=m" (ret)
		: /* no input */
		: "eax", "ecx"
	);
	if (ret)
		goto is486;
	
	/* get vendor info */
	asm volatile(
		"xorl %%eax, %%eax\n\t"
		"cpuid\n\t"
		"movl %%eax, %0\n\t"
		"movl %%ebx, %1\n\t"
		"movl %%edx, %2\n\t"
		"movl %%ecx, %3\n\t"
		: "=m" (boot_cpu_data.cpuid_level),
		  "=m" (boot_cpu_data.x86_vendor_id[0]),
		  "=m" (boot_cpu_data.x86_vendor_id[4]),
		  "=m" (boot_cpu_data.x86_vendor_id[8])
		: /* no input */
		: "eax", "ebx", "ecx", "edx"
	);
	/* do we have processor info as well? */
	if (boot_cpu_data.cpuid_level) {
		asm volatile(
			"movl $1, %%eax\n\t"
			"cpuid\n\t"
			"movb %%al, %%cl\n\t"	/* save reg for future use */
			"andb $0x0f, %%ah\n\t"	/* mask processor family */
			"movb %%ah, %0\n\t"
			"andb $0xf0, %%al\n\t"	/* mask model */
			"shrb $4, %%al\n\t"
			"movb %%al, %1\n\t"
			"andb $0x0f, %%cl\n\t"	/* mask mask revision */
			"movb %%cl, %2\n\t"
			"movl %%edx, %3\n\t"
			: "=m" (boot_cpu_data.x86),
			  "=m" (boot_cpu_data.x86_model),
			  "=m" (boot_cpu_data.x86_mask),
			  "=m" (boot_cpu_data.x86_capability)
			: /* no input */
			: "eax", "al", "ah", "cl", "ch", "edx"
		);
	}
is486:
is386:
	check_x87();

	asm (
		"pushl $idt_descr\n\t"
		"call umcpu_lidt\n\t"
		"addl $4, %esp\n\t"
	);

	if (!ap) {
	  init_video();
	  start_kernel();
	  /*NOTREACHED*/
	} else {
#ifdef CONFIG_SMP
		start_secondary();
#else
		panic("head.c: reached unreachable stuff =;-(");
#endif	/* CONFIG_SMP */
		/*NOTREACHED*/
	}
}

static void check_x87(void)
{
	boot_cpu_data.hard_math = 1;
}

/*
 * The interrupt descriptor table has room for 256 idt's,
 * the global descriptor table is dependent on the number
 * of tasks we can have..
 */
#define IDT_ENTRIES     256
#define GDT_ENTRIES     (12+2*NR_TASKS)

asm (
	".text\n\t"
	".globl idt\n\t"
	".globl gdt\n\t"

	".align\n\t"
	".word 0\n\t"
"idt_descr:\n\t"
	".word 256*8-1\n\t"     /* idt contains 256 entries */
"idt:\n\t"
	".long idt_table\n\t"
"gdt_descr:\n\t"
	".word 0*8-1\n\t"	/* gdt contains ? entries FIXME */
"gdt:\n\t"
	".long gdt_table\n\t"
);

/*
 * This is initialized to create a identity-mapping at 0-4M (for bootup
 * purposes) and another mapping of the 0-4M area at virtual address
 * PAGE_OFFSET.
 */
asm (
".text\n\t"
".org 0x1000\n\t"
"swapper_pg_dir: .globl swapper_pg_dir\n\t"
	/* cannot use macros in inline assembler code? */
	/* => set values later VOSSI */ \
	".fill 1024,4,0\n\t"
);

/*
 * The page tables are initialized to only 4MB here - the final page
 * tables are set up later depending on memory size. The "007" at the
 * end doesn't mean with right to kill, but PRESENT+RW+USER
 */
asm (
".text\n\t"
".org 0x2000\n\t"
"pg0: .globl pg0\n\t"
);

asm (
".text\n\t"
".org 0x3000\n\t"
"pg1: .globl pg1\n\t"
);

asm (
".text\n\t"
".org 0x4000\n\t"
"empty_zero_page: .globl empty_zero_page\n\t"
);

asm (
".text\n\t"
".org 0x5000\n\t"
);

/* FIXME HSHOEXER: ENTRY(stext/_stext)... */

/*
 * This starts the data section. Note that the above is all
 * in the text section because it has alignment requirements
 * that we cannot fulfill any other way.
 */
asm (
".data\n\t"

".align\n\t"                                                                    /*
 * This contains typically 140 quadwords, depending on NR_CPUS.
 *
 * NOTE! Make sure the gdt descriptor in head.S matches this if you
 * change anything.
 */
".globl gdt_table\n\t"
"gdt_table:"
	".quad 0x0000000000000000\n\t"	/* NULL descriptor */
	/* FIXME VOSSI */
);
