/*
 * $Id: umfloppy.c,v 1.1.1.1 2002/08/23 13:52:15 kingst Exp $
 */

#include <linux/assert.h>

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/fd.h>

#include <asm/lotol.h>
#include <asm/uaccess.h>

#define MAJOR_NR	FLOPPY_MAJOR
#include <linux/blk.h>

#include <asm/lotol_fdesc.h>


#define NUM_UMFD	2
#define UMFD_SIZE	512

static int umfd_hardsec[NUM_UMFD];	/* size of real blocks in bytes */
static int umfd_blocksizes[NUM_UMFD];	/* size of 1024 byte blocks :) */
static int umfd_kbsize[NUM_UMFD];	/* size in blocks of 1024 byte */

static int umfd_diskChange[NUM_UMFD];
static int umfd_opencount[NUM_UMFD];

static int umfdID = MAJOR_NR;

/* schedules motor off */
static void floppy_off(unsigned int drive)
{
	/* nothing to do... */
}

#if 0	/* FIXME HSHOEXER */
static void 
do_fd_request(request_queue_t *q) 
{
	unsigned int minor;
	unsigned long offset, len;
	char path[16];
	int fd;
	int ret;
	int request_ok = 1 ;


repeat:
	if (!CURRENT)
		return;

	INIT_REQUEST;

	minor = MINOR(CURRENT->rq_dev);

	if (minor > NUM_UMFD) {
		end_request(0);
		goto repeat;
	}

	if ((CURRENT->sector + CURRENT->current_nr_sectors)
	    > umfd_kbsize[minor] * 2) {
		end_request(0);
		goto repeat;
	}

	offset = CURRENT->sector << 9;
	len = CURRENT->current_nr_sectors << 9;

	CURRENT->sector += CURRENT->current_nr_sectors;
	CURRENT->nr_sectors -= CURRENT->current_nr_sectors;

	sprintf(path, "fd%d/media", minor);
	fd = FD_FD_MEDIA(minor);
	/*
	fd = lotol_open(path,
		CURRENT->cmd == READ ? LOTOL_O_RDONLY : LOTOL_O_WRONLY, 0);
	*/
	assert(0 <= fd);

	ret = lotol_lseek(fd, offset, 0);
	assert(ret == offset);

	if (CURRENT->cmd == READ)
		ret = lotol_read(fd, CURRENT->buffer, len);
	else
		ret = lotol_write(fd, CURRENT->buffer, len);
	/* don't do the following assert, because it will panic the kernel every
	 * time a block device fault is injected and affects this read! */
	/* assert(ret == len);  */
	if(ret != len) {
		request_ok = 0 ;
		printk ("umfloppy: cmd %d failed\n", CURRENT->cmd);
	}

	/*
	close will be done by lilo
	*/

	end_request(request_ok);
	goto repeat;
}
#endif /* FIXME HSHOEXER */


/* perform ioctl on drive to get floppy size */
int umfd_get_fdsize(kdev_t rdev)
{
	int fd;
	int ret;
	int size;
	unsigned int minor;
	char path[16];
	struct floppy_struct geom;

	minor = MINOR(rdev);
	sprintf(path, "fd%d/media", minor);
	fd = FD_FD_MEDIA(minor);

	/*
	fd = lotol_open(path, LOTOL_O_RDONLY, 0);
	*/
	if (fd < 0) return 0;

	ret = lotol_ioctl(fd, FDGETPRM, (long)&geom);

	/* we tried an ioctl on a real file... */
	if (ret < 0) {
		ret = lotol_lseek(fd, 0, 2);
		if (ret < 0) size = 0;
		else size = ret / BLOCK_SIZE;
	} else 
		/* floppy_struct.size is in 512 byte blocks, we need 1024 */
		size = geom.size / 2;
	
	blk_size[MAJOR_NR][minor] = size;
	umfd_kbsize[minor] = size;

	return size;
}


void umfd_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;

	/* only unopened devices are marked as changed */
	if ( *(int *)dev_id == umfdID) {
		for (i = 0; i < NUM_UMFD; i++) {
			if (umfd_opencount[i] == 0) umfd_diskChange[i] = 1;
		}
	}
}


static int umfd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor;

	if (!inode || !inode->i_rdev)
		return -EINVAL;

	minor = MINOR(inode->i_rdev);

	switch (cmd) {
	default:
		return -EINVAL;
	};

	return 0;
}


static int umfd_open(struct inode * inode, struct file * filp)
{
	if (umfd_get_fdsize(inode->i_rdev) == 0) return -EIO;
	umfd_opencount[MINOR(inode->i_rdev)]++;
	return 0;
}


static int umfd_release(struct inode * inode, struct file * filp)
{
	umfd_opencount[MINOR(inode->i_rdev)]--;
	assert(umfd_opencount[MINOR(inode->i_rdev)] >= 0);
	return 0;
}


static int umfd_check_media_change(kdev_t dev)
{
	if (umfd_diskChange[MINOR(dev)]) {
		umfd_diskChange[MINOR(dev)] = 0;
		return 1;
	} else return 0;
}

static struct block_device_operations umfd_fops = {
	open:			umfd_open,
	release:		umfd_release,
	ioctl:			umfd_ioctl,
	check_media_change:	umfd_check_media_change,
	revalidate:		NULL,
};

#if 0	/* FIXME HSHOEXER */
static struct file_operations umfd_fops = {
	NULL,				/* lseek - default */
	block_read,			/* read - block dev read */
	block_write,			/* write - block dev read */
	NULL,				/* readdir - not here */
	NULL,				/* poll */
	umfd_ioctl,			/* ioctl */
	NULL,				/* mmap */
	umfd_open,			/* open */
	NULL,				/* flush */
	umfd_release,			/* release */
	block_fsync,			/* fsync */
	NULL,				/* fasync */
	umfd_check_media_change,	/* media change */
	NULL				/* revalidate */
};
#endif

int __init floppy_init(void) {

	int i;

	printk("UM floppy init begin\n") ;
	if (register_blkdev(MAJOR_NR, "umfd", &umfd_fops)) {
		printk("UMFD: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	/* require 'interrupt' */
	i = request_irq(LOTOL_SIGHUP + 16, umfd_interrupt, SA_SHIRQ, "umfd", 
	    &umfdID);
	if (i < 0) panic("umfd_init: request_irq");

#if 0	/* FIXME HSHOEXER */
	blk_dev[MAJOR_NR].request_fn = &do_fd_request;
#endif

	printk("UM floppy driver initialized (major=%d):\n", MAJOR_NR);
	for (i = 0; i < NUM_UMFD; i++) {
		char path[16];
		int fd;
		int ret;

		sprintf(path, "fd%d/media", i);
		fd = FD_FD_MEDIA(i);
		ret = lotol_lseek(fd, 0, 2);
		if (ret < 0) {
			umfd_hardsec[i] = 0;
			umfd_blocksizes[i] = 0;
			printk("    %s: not found\n", path);
		} else {
			umfd_hardsec[i] = UMFD_SIZE;
			umfd_blocksizes[i] = BLOCK_SIZE;
			/*
			lilo does close
			*/
			printk("    %s: device found (%d byte per block)\n", 
			    path, umfd_blocksizes[i]);
		}
	
		/* depends on current media, so default to 0 */	
		umfd_kbsize[i] = 0;

		umfd_opencount[i] = 0;
		umfd_diskChange[i] = 0;
	}

	hardsect_size[MAJOR_NR] = umfd_hardsec;
	blksize_size[MAJOR_NR] = umfd_blocksizes;
	blk_size[MAJOR_NR] = umfd_kbsize;

	printk("UM floppy init end\n") ;
	return 0;
}

#ifdef MODULE

/* ... */
                                                                                #else
/* eject the boot floppy (if we need the drive for a different root floppy) */
/* This should only be called at boot time when we're sure that there's no
 * resource contention. */
void floppy_eject(void)
{
	/* FIXME VOSSI */
}
#endif
