#ifndef _ASM_LOTOL_H
#define _ASM_LOTOL_H

extern int errno;
#include <linux/types.h>
#define __KERNEL_SYSCALLS__
#include <asm/unistd.h>

/* GWD */
#define __NR_ztrace		233

/*
 * constants
 */
/* <asm/errno.h> */
#define LOTOL_EINTR		4
#define LOTOL_EIO		5
#define LOTOL_EAGAIN		11
#define LOTOL_ENOSYS		38
#define LOTOL_ECONNREFUSED	111

/* <asm/fcntl.h> */
#define LOTOL_O_RDONLY		00
#define LOTOL_O_WRONLY		01
#define LOTOL_O_RDWR		02
#define LOTOL_O_CREAT		0100
#define LOTOL_O_EXCL		0200
#define LOTOL_O_NOCTTY		0400
#define LOTOL_O_TRUNC		01000
#define LOTOL_O_APPEND		02000
#define LOTOL_O_NONBLOCK	04000
#define LOTOL_O_ASYNC		020000

#define LOTOL_F_GETFL		3
#define LOTOL_F_SETFL		4
#define LOTOL_F_SETOWN		8
#define LOTOL_F_GETOWN		9

/* <asm/ioctls.h> */
#define LOTOL_TCGETS		0x5401
#define LOTOL_TCSETS		0x5402

/* <asm/mman.h> */
#define LOTOL_PROT_READ		0x1
#define LOTOL_PROT_WRITE	0x2
#define LOTOL_PROT_EXEC		0x4

#define LOTOL_MAP_SHARED	0x01
#define LOTOL_MAP_FIXED		0x10

/* <asm/ptrace.h> */
#define LOTOL_PTRACE_PEEKDATA	 2
#define LOTOL_PTRACE_SINGLESTEP	 9
#define LOTOL_PTRACE_GETREGS	12
#define LOTOL_PTRACE_SETREGS	13

/* <asm/signal.h> */

#define LOTOL__NSIG		64

#define LOTOL_SIGHUP		 1
#define LOTOL_SIGILL		 4
#define LOTOL_SIGTRAP		 5
#define LOTOL_SIGBUS		 7
#define LOTOL_SIGFPE		 8
#define LOTOL_SIGUSR1		10
#define LOTOL_SIGSEGV		11
#define LOTOL_SIGUSR2		12
#define LOTOL_SIGPIPE		13
#define LOTOL_SIGALRM		14
#define LOTOL_SIGURG		23
#define LOTOL_SIGVTALRM		26
#define LOTOL_SIGIO		29

#define LOTOL_SIG_BLOCK		0
#define LOTOL_SIG_UNBLOCK	1
#define LOTOL_SIG_SETMASK	2

#define LOTOL_SA_SIGINFO	0x00000004
#define LOTOL_SA_RESTART	0x10000000
#define LOTOL_SA_ONSTACK	0x08000000
#define LOTOL_SA_NODEFER	0x40000000

/* <asm/termbits.h> */
#define LOTOL_ISIG	0000001
#define LOTOL_ICANON	0000002
#define LOTOL_XCASE	0000004
#define LOTOL_ECHO	0000010
#define LOTOL_ECHOE	0000020
#define LOTOL_ECHOK	0000040
#define LOTOL_ECHONL	0000100
#define LOTOL_NOFLSH	0000200
#define LOTOL_TOSTOP	0000400
#define LOTOL_ECHOCTL	0001000
#define LOTOL_ECHOPRT	0002000
#define LOTOL_ECHOKE	0004000
#define LOTOL_FLUSHO	0010000
#define LOTOL_PENDIN	0040000
#define LOTOL_IEXTEN	0100000

/* <asm/wait.h> */
#define LOTOL_WUNTRACED	0x00000002

/* <linux/ptrace.h> */
#define LOTOL_PTRACE_TRACEME	0
#define LOTOL_PTRACE_SYSCALL	24

/* <linux/time.h> */
#define LOTOL_ITIMER_REAL	0
#define LOTOL_ITIMER_VIRTUAL	1

/* netinet/in.h */
#define LOTOL_PF_INET		2
#define LOTOL_AF_INET		2
#define LOTOL_SOCK_STREAM	1
#define LOTOL_SOCK_DGRAM	2
#define LOTOL_INADDR_ANY	0

/* <sys/wait.h> */
#define LOTOL_WIFEXITED(x)	(((x) & 0x7f) == 0)
#define LOTOL_WIFSTOPPED(x)	(((x) & 0x7f) == 0x7f)
#define LOTOL_WIFSIGNALED(x)	(((x) & 0x7f) != 0 && ((x) & 0x7f) != 0x7f)
#define LOTOL_WSTOPSIG(x)	(((x) >> 8) & 0xff)
#define LOTOL_WEXITSTATUS(x)	(((x) >> 8) & 0xff)
#define LOTOL_WTERMSIG(x)	((x) & 0x7f)

/*
 * types
 */
/* asm/signal.h */
typedef struct {
	unsigned long sig[16];
} lotol_sigset_t;

struct lotol_sigaction {
	void (*sa_sigaction)(int, void *, void *);
	unsigned long sa_flags;
	void (*sa_restorer)(void);
	lotol_sigset_t sa_mask;
};

typedef struct lotol_stack_t {
	void *ss_sp;
	int ss_flags;
	size_t ss_size;
} lotol_stack_t;

/* <asm/termbits.h> */
#define LOTOL_NCCS	19
struct lotol_termios {
	unsigned int c_iflag;
	unsigned int c_oflag;
	unsigned int c_cflag;
	unsigned int c_lflag;
	unsigned char c_line;
	unsigned char c_cc[LOTOL_NCCS];
};

/* <asm/user.h> */
struct lotol_user_regs_struct {
	long ebx, ecx, edx, esi, edi, ebp, eax;
	unsigned short ds, __ds, es, __es;
	unsigned short fs, __fs, gs, __gs;
	long orig_eax, eip;
	unsigned short cs, __cs;
	long eflags, esp;
	unsigned short ss, __ss;
};

/* <linux/resource.h> */
struct lotol_rusage {
	int dummy;	/* not used */
};

/* <linux/socket.h> */
struct lotol_sockaddr {
	unsigned short sa_family;
	char sa_data[14];
};

/* <linux/time.h> */
struct lotol_timeval {
	unsigned long tv_sec;
	unsigned long tv_usec;
};

struct lotol_timezone {
	int tz_minuteswest;
	int tz_dsttime;
};

struct lotol_itimerval {
	struct lotol_timeval it_interval;
	struct lotol_timeval it_value;
};

/* netinet/in.h */
struct lotol_sockaddr_in {
	short sin_family;
	short sin_port;
	long sin_addr;
	char dummy[8];
};

/*
 * variables
 */
#undef errno
extern int errno;

/*
 * functions
 */
#define _lotol_syscall_return(type, res) \
do { \
	if ((unsigned long)(res) >= (unsigned long)(-125)) { \
		errno = -(res); \
		res = -1; \
	} \
	return (type) (res); \
} while (0)

#define _lotol_syscall0(type,name) \
type lotol_##name(void) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define _lotol_syscall1(type,name,type1,arg1) \
type lotol_##name(type1 arg1) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define _lotol_syscall2(type,name,type1,arg1,type2,arg2) \
type lotol_##name(type1 arg1,type2 arg2) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)),"c" ((long)(arg2)), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define _lotol_syscall3(type,name,type1,arg1,type2,arg2,type3,arg3) \
type lotol_##name(type1 arg1,type2 arg2,type3 arg3) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)),"c" ((long)(arg2)), \
		  "d" ((long)(arg3)), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define _lotol_syscall4(type,name,type1,arg1,type2,arg2,type3,arg3,type4,arg4) \
type lotol_##name (type1 arg1, type2 arg2, type3 arg3, type4 arg4) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)),"c" ((long)(arg2)), \
		  "d" ((long)(arg3)),"S" ((long)(arg4)), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define _lotol_syscall5(type,name,type1,arg1,type2,arg2,type3,arg3,type4,arg4,type5,arg5) \
type lotol_##name (type1 arg1, type2 arg2, type3 arg3, type4 arg4, type5 arg5) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)),"c" ((long)(arg2)), \
		  "d" ((long)(arg3)),"S" ((long)(arg4)), \
		  "D" ((__NR_##name << 16) | (long)(arg5))); \
	_lotol_syscall_return(type,__res); \
}

static inline _lotol_syscall2(int,socketcall,int,nr,void *,arg);
static inline int lotol_bind(int fd, struct lotol_sockaddr *addr, int addrlen)
{
	struct {
		int fd;
		struct lotol_sockaddr *addr;
		int addrlen;
	} args;

	args.fd = fd;
	args.addr = addr;
	args.addrlen = addrlen;
	return lotol_socketcall(2, &args);
}
static inline _lotol_syscall1(int,close,int,fd);
static inline int lotol_connect(int fd, struct lotol_sockaddr *addr, int addrlen)
{
	struct {
		int fd;
		struct lotol_sockaddr *addr;
		int addrlen;
	} args;

	args.fd = fd;
	args.addr = addr;
	args.addrlen = addrlen;
	return lotol_socketcall(3, &args);
}
#define __NR__exit __NR_exit
static inline _lotol_syscall1(void,_exit,int,retval);
static inline _lotol_syscall3(int,fcntl,int,fd,int,cmd,long,arg);
static inline _lotol_syscall0(int,getpid);
static inline _lotol_syscall0(int,sched_yield);
static inline _lotol_syscall2(int,gettimeofday,struct lotol_timeval *,tv,struct lotol_timezone *,tz);
static inline _lotol_syscall3(int,ioctl,int,fd,int,request,long,arg);
static inline _lotol_syscall2(int,kill,int,pid,int,sig);
static inline _lotol_syscall3(int,lseek,int,fd,int,offset,int,whence);
static inline _lotol_syscall2(int,ztrace,int,command,int,mark);
static inline _lotol_syscall5(int,_llseek,uint,fd,ulong,hu,ulong,lo,loff_t *,res,uint,wh);
#define __NR__mmap __NR_mmap
static inline _lotol_syscall1(int, _mmap, void *, args);
static inline void *lotol_mmap(void *addr, unsigned long len, int prot, int flags, int fd, unsigned long offset)
{
	struct mmap_arg_struct {
		unsigned long addr;
		unsigned long len;
		unsigned long prot;
		unsigned long flags;
		unsigned long fd;
		unsigned long offset;
	} args;

	args.addr = (unsigned long) addr;
	args.len = len;
	args.prot = prot;
	args.flags = flags;
	args.fd = fd;
	args.offset = offset;
	return (void *) lotol__mmap(&args);
}
static inline _lotol_syscall3(int, mprotect, void *, addr, unsigned long, len, int, prot);
static inline _lotol_syscall2(int, munmap, void *, addr, unsigned long, len);
static inline _lotol_syscall3(int,open,const char *,name,int,flags,int,mode);
static inline _lotol_syscall4(int,rt_sigaction,int,signum,const struct lotol_sigaction *,act,struct lotol_sigaction *,oldact,int,sigsetsize);
static inline _lotol_syscall2(int,rt_sigsuspend,lotol_sigset_t *,unewset,size_t,sigsetsize);
static inline _lotol_syscall3(int,read,int,fd,void *,buf,int,count);
static inline int lotol_recvfrom(int s, void *buf, int len, int flags, struct lotol_sockaddr *from, int *fromlen)
{
	struct {
		int s;
		void *buf;
		int len;
		int flags;
		struct lotol_sockaddr *from;
		int *fromlen;
	} args;

	args.s = s;
	args.buf = buf;
	args.len = len;
	args.flags = flags;
	args.from = from;
	args.fromlen = fromlen;
	return lotol_socketcall(12, &args);
}
static inline int lotol_sendto(int s, void *buf, int len, int flags, struct lotol_sockaddr *to, int tolen)
{
	struct {
		int s;
		void *buf;
		int len;
		int flags;
		struct lotol_sockaddr *to;
		int tolen;
	} args;

	args.s = s;
	args.buf = buf;
	args.len = len;
	args.flags = flags;
	args.to = to;
	args.tolen = tolen;
	return lotol_socketcall(11, &args);
}
static inline _lotol_syscall3(int,setitimer,int,which,const struct lotol_itimerval *,val,struct lotol_itimerval *,oldval);
static inline _lotol_syscall3(int,sigaction,int,signum,const struct lotol_sigaction *,act,struct lotol_sigaction *,oldact);
static inline _lotol_syscall2(int,sigaltstack,lotol_stack_t *,ss,lotol_stack_t *,oss);
static inline _lotol_syscall3(int,sigprocmask,int,how,lotol_sigset_t *,set,lotol_sigset_t *,oldset);
static inline int lotol_socket(int domain, int type, int protocol)
{
	struct {
		int domain;
		int type;
		int protocol;
	} args;

	args.domain = domain;
	args.type = type;
	args.protocol = protocol;
	return lotol_socketcall(1, &args);
}
static inline _lotol_syscall3(int,write,int,fd,const void *,buf,int,count);

static inline void lotol_sigaddset(lotol_sigset_t *set, int _sig)
{
	unsigned long sig = _sig - 1;
	set->sig[sig / 32] |= 1 << (sig % 32);
}

static inline void lotol_sigdelset(lotol_sigset_t *set, int _sig)
{
	unsigned long sig = _sig - 1;
	set->sig[sig / 32] &= ~(1 << (sig % 32));
}

static inline void lotol_sigemptyset(lotol_sigset_t *set)
{
	long *s;
	int n;

	for (s = set->sig, n = 0; n < sizeof(*set); s++, n += sizeof(*s))
		*s = 0;
}

static inline void lotol_sigfillset(lotol_sigset_t *set)
{
	long *s;
	int n;

	for (s = set->sig, n = 0; n < sizeof(*set); s++, n += sizeof(*s))
		*s = 0xffffffff;
}

static inline int lotol_sigismember(lotol_sigset_t *set, int _sig)
{
	unsigned long sig = _sig - 1;
	return 1 & (set->sig[sig / 32] >> (sig % 32));
}

static inline unsigned short lotol_htons(unsigned short x)
{
	return ((x & 0x00ff) << 8)
	     | ((x & 0xff00) >> 8);
}

static inline unsigned short lotol_ntohs(unsigned short x)
{
	return lotol_htons(x);
}

static inline unsigned long lotol_htonl(unsigned long x)
{
	return ((x & 0x000000ff) << 24)
	     | ((x & 0x0000ff00) <<  8)
	     | ((x & 0x00ff0000) >>  8)
	     | ((x & 0xff000000) >> 24);
}

static inline unsigned long lotol_ntohl(unsigned long x)
{
	return lotol_htonl(x);
}

// STK these are for fast context switching
#define __NR_getguestpgd  235
static inline _lotol_syscall2(int,getguestpgd,int,cmd,unsigned long,pgdIdx);

#define __NR_switchguest  236
static inline _lotol_syscall2(int,switchguest,unsigned long,prevIdx,
                              unsigned long,nextIdx);
#define __NR_syncprint 237
static inline _lotol_syscall1(int,syncprint,char *,buf);

#endif /* _ASM_LOTOL_H */
